# Nanolith

Multithreading in minutes. _(More intuitive and feature-rich than [Piscina](https://www.npmjs.com/package/piscina)!)_

[![TypeScript](https://badgen.net/badge/-/TypeScript/blue?icon=typescript&label)](https://www.typescriptlang.org/) [![CircleCI](https://circleci.com/gh/mstephen19/nanolith.svg?style=svg)](https://app.circleci.com/pipelines/github/mstephen19/nanolith) [![Install size](https://packagephobia.com/badge?p=nanolith@latest)](https://packagephobia.com/result?p=nanolith@latest)

[![Version](https://img.shields.io/npm/v/nanolith?color=blue)](https://github.com/mstephen19/nanolith/releases) ![Weekly downloads](https://img.shields.io/npm/dw/nanolith?color=violet) ![Libraries.io dependency status](https://img.shields.io/librariesio/release/npm/nanolith) [![GitHub issues](https://img.shields.io/github/issues/mstephen19/nanolith?color=red)](https://github.com/mstephen19/nanolith/issues)

<center>
    <img src="https://user-images.githubusercontent.com/87805115/199340985-d76cc3ea-6abb-4a4e-ac1b-a95fc693947f.png" width="550">
</center>

## ❔ About

✨**Nanolith**✨ is a performant, reliable, easy-to-use, and well-documented multithreading library. It serves to not only build upon, but entirely replace the _(deprecated)_ [Threadz](https://github.com/mstephen19/threadz) library.

There have always been three main goals for Nanolith:

1. Performance 🏃
2. Ease-of-use 😇
3. Seamless TypeScript support 😎

### So what can you do with it?

Here's a quick rundown of everything you can do in Nanolith:

- Offload expensive tasks to separate threads.
- Spawn up separate-threaded "nanoservices" that can run any tasks you want.
- Communicate back and forth between threads by sending messages.
- Stream data between threads with the already familiar [`node:stream`](https://nodejs.org/api/stream.html) API.
- Share memory between threads using the familiar-feeling `SharedMap` class.

## 📖 Table of contents

- [❔ About](#-about)
- [💾 Installation](#-installation)
- [📝 Defining your tasks](#-defining-your-tasks)
  - [`define()` options](#define-options)
- [👷 Running a task](#-running-a-task)
  - [Task function options](#task-function-options)
- [🎩 Understanding services](#-understanding-services)
  - [`launchService()` options](#launchservice-options)
- [🎬 Coordinating services](#-coordinating-services)
- [🪝 Hooks](#-hooks)
- [🚨 Managing concurrency](#-managing-concurrency)
- [📨 Communicating between threads](#-communicating-between-threads)
  - [Between a service and the main thread](#between-a-service-and-the-main-thread)
  - [Between all threads](#between-all-threads)
- [📡 Streaming data between threads](#-streaming-data-between-threads)
- [💾 Sharing memory between threads](#-sharing-memory-between-threads)
- [🧑‍🏫 Examples](#-examples)
- [📜 License](#-license)

## 💾 Installation

The latest version can be installed via any package manager of your choice.

```shell
npm install nanolith@latest
# or
yarn add nanolith@latest
```

Beta versions are released under the **next** tag and can be installed like this:

```shell
npm install nanolith@next
# or
yarn add nanolith@next
```

## 📝 Defining your tasks

A **task** is any function that **you** define which is accessible by Nanolith's APIs. Tasks can be defined using the `define()` function in a separate file dedicated to definitions.

```TypeScript
// worker.ts 💼
import { define } from 'nanolith';

// Exporting the variable is not a requirement, but it is
// necessary to somehow export the resolved value of the
// function in order to have access to it later on.
export const worker = await define({
    add(x: number, y: number) {
        return x + y;
    },
    async waitThenAdd(x: number, y: number) {
        await new Promise((resolve) => setTimeout(resolve, 5e3))
        return x + y;
    },
    // Functions don't have to be directly defined within the
    // object, they can be defined elsewhere outside, or even
    // imported from a totally different module.
    subtract,
});

function subtract(x: number, y: number) {
    return x - y;
};
```

By passing functions into `define()`, you immediately turn them into multithreadable **tasks**. No further configuration is required.

### `define()` options

As seen above, the first argument to `define()` is an object containing your functions. The second parameter is an object accepting the following _(optional)_ configurations:

| Name | Type | About |
|-|-|-|
| `file` | **string** | If `define()`'s file location detection is not working correctly, the true file location for the set of definitions can be provided here. |
| `identifier` | **string** | A unique identifier for the set of definitions. Overrides the auto-identifier generated by Nanolith. |
| `safeMode` | **boolean** | Whether or not to prevent the usage of the returned **Nanolith** API from within the same file where their definitions were created. Defaults to `true`. |

## 👷 Running a task

After [defining](#-defining-your-tasks) a set of tasks, you can import them and call them anywhere by directly using the **Nanolith** API resolved by the `define()` function. The only difference is that instead of being called on the main thread, a new thread will be created for the task and it will be run there.

```TypeScript
// 💡 index.ts
// Importing the Nanolith API we created in worker.ts
import { worker } from './worker.js';

// Run the "add" function on a separate thread and wait
// for it to complete before moving forward.
const result = await worker({
    // Provide the name of the task.
    name: 'add',
    // Provide the parameters of the function.
    params: [2, 3],
});

// The result is sent back to the main thread
// and resolved by the task function call.
console.log(result); // -> 5
```

The new thread's process is shut down after the task finishes.

> **📝 Note:** Notice that even with the synchronous `add()` function, it is now asynchronous when being multithreaded.

### Task function options

`name` and `params` are amongst many of the possible options that can be passed in when running a task:

| Name | Type | About |
|-|-|-|
| `name` | **string** | The name of the task to call. Must be present on the set of definitions. |
| `params` | **any[]** | The arguments for the task in array form. |
| `priority` | **boolean** | Whether or not to treat the task's worker as priority over others when being queued into the `pool`. |
| `reffed` | **boolean** | When `true`, the underlying `Worker` instance is [reffed](https://nodejs.org/api/worker_threads.html#workerref). Defaults to `false`. |
| `messengers` | [**Messenger**](#between-all-threads)**[]** | The `Messenger`s that should be accessible to the task. |
| `options` | **object** | An object containing _most_ of the options available on the [`Worker` constructor](https://nodejs.org/api/worker_threads.html#new-workerfilename-options). |

## 🎩 Understanding services

**Services** are Nanolith's flagship feature. Running a task on a service works similarly to [running a task](#-running-a-task) normally; however, the key difference is that the thread only shuts down when you tell it to. This means that you can run multiple tasks on the same thread rather than spawning up a new one for each call.

Considering the definitions we created [here](#-defining-your-tasks), here is how a service would be launched and a task would be called on it.

```TypeScript
// 💡 index.ts
// Importing the Nanolith API we created in worker.ts
import { worker } from './worker.js';

// Spawn up a new thread that has access to all of
// our tasks.
const service = await worker.launchService();

// Command the service thread to run the "add" function.
const result = await service.call({
    name: 'waitThenAdd',
    params: [2, 3],
});

// We can run service.call() as many times as we want, and
// all those tasks will be called on the same thread...

// Similarly to regular task calls, the return value
// is sent back to the main thread and resolve by the call.
console.log(result);

// Shut down the second thread.
await service.close();
```

### `launchService()` options

The configurations for `Nanolith.launchService()` are nearly identical to the [task function options](#task-function-options) with the addition of `exceptionHandler`:

| Name | Type | About |
|-|-|-|
| `exceptionHandler` | **function** | An optional but _highly recommended_ option that allows you to catch uncaught exceptions within the service. |
| `priority` | **boolean** | Whether or not to treat the service's worker as priority over others when being queued into the `pool`. |
| `reffed` | **boolean** | When `true`, the underlying `Worker` instance is [reffed](https://nodejs.org/api/worker_threads.html#workerref). Defaults to `false`. |
| `messengers` | [**Messenger**](#between-all-threads)**[]** | The `Messenger`s that should be accessible to the service. |
| `options` | **object** | An object containing _most_ of the options available on the [`Worker` constructor](https://nodejs.org/api/worker_threads.html#new-workerfilename-options). |

<!-- todo: Go over all methods & properties available on Service -->

## 🎬 Coordinating services

In a scalable application utilizing multiple identical [services](#launchservice-options), it is possible to optimize them by treating the main thread as an orchestrator and managing the workloads on each service. Nanolith's `ServiceCluster` automatically does this for you.

```TypeScript
// 💡 index.ts
// Importing the Nanolith API we created in worker.ts
import { worker } from './worker.js';

// Launch 6 identical services at the same time.
// Returns a "ServiceCluster" instance.
const cluster = await worker.clusterize(6, {
    // These options will be applied to all of the 6
    // services being launched.
    exceptionHandler({ error, terminate }) {
        console.error(error);
    },
    priority: true;
});

// Use the least busy service on the cluster.
// This is the service that is currently running
// the least amount of task calls.
const service = cluster.use();

// Call the task on the service as you normally would.
const result = await service.call({
    name: 'subtract',
    params: [10, 5],
});

console.log(result);

// Close all services on the cluster.
await cluster.closeAll();
```

For simplicity of the above example, we are only running a single task. However, `ServiceCluster` can be used to run a large amount of heavy operations in true parallel on multiple services.

<!-- todo: Go over all methods & properties available on ServiceCluster -->

## 🪝 Hooks

For a bit of finer control over your services and tasks, three hooks are available and can be provided directly to [`define()`](#-defining-your-tasks).

```TypeScript
// worker.ts 💼
import { define } from 'nanolith';

export const worker = await define({
    // Runs before a service is launched and before the
    // "launchService" function resolves its promise.
    __initializeService(threadId) {
        console.log(`Initializing service on thread: ${threadId}`);
    },
    // Runs before a task is called.
    __beforeTask({ name, inService }) {
        console.log(`Running task ${name}.`);
        // You have access to "inService", which tells you if the
        // task will run standalone, or within a service.
        console.log(`${inService ? 'Is' : 'Is not'} in a service.`);
    },
    // Runs after a task is called.
    __afterTask({ name, inService }) {
        console.log(`Finished task ${name}`);
    },
    // Define your tasks here...
});
```

These hooks run on the same thread as their service/task.

## 🚨 Managing concurrency

Nanolith automatically manages the concurrency your services and task calls with the internal `pool` class. By default, the maximum concurrency is one thread per core on the machine. This is a safe value to go with; however, the `maxConcurrency` can be modified up using one of the set `ConcurrencyOption`s.

```TypeScript
// index.ts 💡
// Importing the pool
import { pool, ConcurrencyOption } from 'nanolith';

// One thread per four cores.
pool.setConcurrency(ConcurrencyOption.Quarter);
// One thread per two cores.
pool.setConcurrency(ConcurrencyOption.Half);
// Default concurrency. One thread per core (x1).
pool.setConcurrency(ConcurrencyOption.Default);
// One thread per core.
pool.setConcurrency(ConcurrencyOption.x1);
// Two threads per core.
pool.setConcurrency(ConcurrencyOption.x2);
// Four threads per core.
pool.setConcurrency(ConcurrencyOption.x4);
// Six threads per core.
pool.setConcurrency(ConcurrencyOption.x6);
// Eight threads per core.
pool.setConcurrency(ConcurrencyOption.x8);
// Ten threads per core.
// Warning: This is overkill.
pool.setConcurrency(ConcurrencyOption.x10);
```

<!-- todo: go over the various other properties and methods on the pool -->

## 📨 Communicating between threads

There are two ways of communicating between threads in Nanolith.

### Between a service and the main thread

When using [services](#-understanding-services), you are automatically able to communicate between the service and main thread with no extra work.

The [`__initializeService()` hook](#-hooks) can be used to register listeners on the `MainThread`:

```TypeScript
// worker.ts 💼
import { define, MainThread } from 'nanolith';

export const worker = await define({
    // When the service is launched, this function will be
    // called.
    __initializeService() {
        // Register a listener for a message coming from the
        // main thread.
        MainThread.onMessage<string>((message) => {
            // Log the message when it is received.
            console.log(message);

            // Then, after the message is received, send a
            // confirmation to the main thread.
            MainThread.sendMessage('hello from the service!');
        });
    },
});
```

Then, the `.onMessage()` and `.sendMessage()` methods on the created [service](#-understanding-services) can be used to send messages to and receive messages from the service:

```TypeScript
// 💡 index.ts
import { worker } from './worker.js';

const service = await worker.launchService();

// After the service is launched and initialized, send
// a message to it.
service.sendMessage('hello from the main thread');

// Register a listener for when a message is received
// from the service.
service.onMessage(async (message) => {
    // When a message is received, first log the message's
    // contents.
    console.log(message);

    // Then, close the service.
    await service.close();
});
```

### Between all threads

A bit of extra work is required when there is a need to communicate between all threads (including the main thread). First, an instance of `Messenger` must created. That instance can then be exposed to as many services and tasks as you want.

Within task functions, the `.use()` method on `MessengerList` can be used to grab hold of `Messenger`s exposed to the thread:

```TypeScript
// worker.ts 💼
import { define, MessengerList } from 'nanolith';

export const worker = await define({
    // When the service is launched, this function will be
    // called.
    async __initializeService() {
        // Grab hold of the exposed "foo" messenger.
        const fooMessenger = await MessengerList.use('foo');
        // Register a listener for a message received on the
        // messenger.
        fooMessenger.onMessage<string>((message) => {
            // Log the message when it is received.
            console.log(message);
            // Exit the process once any message has been received.
            process.exit();
        });
    },
    // A task function which will trigger a message to be sent
    // on the exposed "foo" messenger.
    async sendSomeMessage() {
        const fooMessenger = await MessengerList.use('foo');
        fooMessenger.sendMessage('hello from other service!');
    },
});
```

The messenger instance can be exposed to a [task call](#task-function-options) or [service](#launchservice-options) by using the `messengers` option.

```TypeScript
// 💡 index.ts
import { Messenger } from 'nanolith';
import { worker } from './worker.js';

// Create a new messenger with the ID of "foo"
const fooMessenger = new Messenger('foo');

// Launch two services that have the "foo" messenger
// exposed to both of them.
const service1 = await worker.launchService({
    messengers: [fooMessenger],
});
const service2 = await worker.launchService({
    messengers: [fooMessenger],
});

// Call the task which sends a message on the exposed
// "foo" messenger.
await service1.call({ name: 'sendSomeMessage' });

// Finally close the first service. The second will have
// already closed itself.
await service1.close();
```

> **Note:** The `Messenger` class can be used to communicate between all threads. That means between [task calls](#-running-a-task), between [services](#-understanding-services), between the main thread and multiple services/task calls, etc.

## 📡 Streaming data between threads

It's possible to stream data from one thread to another either using [`Service`](#-understanding-services), [`Messenger`](#between-all-threads), and [`MainThread`](#between-a-service-and-the-main-thread). All have the `.createStream()` and `.onStream()` methods.

```TypeScript
// worker.ts 💼
import { define, MainThread } from 'nanolith';
import { createWriteStream } from 'fs';

export const worker = await define({
    __initializeService() {
        // Wait for streams coming from the main thread.
        MainThread.onStream((stream) => {
            const writeStream = createWriteStream('./movie.mp4');
            // Once the stream has finished, notify the main thread
            // that the service is ready to be closed.
            stream.on('end', () => {
                MainThread.sendMessage('close please');
            });

            // Pipe the received stream right into our created
            // write stream.
            stream.pipe(writeStream);
        });
    },
});
```

```TypeScript
// 💡 index.ts
import axios from 'axios';
import { worker } from './worker.js';
import type { Readable } from 'stream';

const service = await worker.launchService();

// Once a message has been received from the service,
// close it immediately.
service.onMessage(async () => {
    await service.close();
});

// Get a Readstream for the entire movie "Edward Scissorhands"
const { data: readStream } = await axios.get<Readable>(
    'https://stream-1-1-ip4.loadshare.org/slice/3/VideoID-qbfnKjG4/CXNa4S/uSDJeP/BXvsDm/Jmrsew/360?name=edward-scissorhands_360&token=ip=85.160.33.237~st=1672263375~exp=1672277775~acl=/*~hmac=de82d742e7cda87859d519fdbf179416d67366497f2e65c103de830b379b1e8b',
    {
        responseType: 'stream',
    }
);

// Send the stream to the service to be handled.
readStream.pipe(await service.createStream());
```

When using `Messenger`, things work a bit differently. The `.onStream()` method takes a different type of callback that must first accept the stream before handling it. This is because with messengers, there are multiple possible recipients, and not all of them might want to accept the stream.

Again, we use the [`__initializeService()` hook](#-hooks):

```TypeScript
// worker.ts 💼
import { define, MessengerList, MainThread } from 'nanolith';
import { createWriteStream } from 'fs';

export const worker = await define({
    async __initializeService() {
        // Use the exposed "foo" messenger
        const fooMessenger = await MessengerList.use('foo');

        // Register a listener for once a stream is received.
        fooMessenger.onStream(({ metaData, accept }) => {
            // If the metadata of the stream matches what we
            // want, we will continue. Otherwise we'll decline
            // the stream by doing nothing.
            if (metaData.scissorhands !== true) return;

            // Retrieve the stream by calling the "accept" function.
            const stream = accept();

            const writeStream = createWriteStream('./movie.mp4');
            // Once the stream has finished, notify the main thread
            // that the service is ready to be closed.
            stream.on('end', () => {
                MainThread.sendMessage('close please');
            });

            // Pipe the received stream right into our created
            // write stream.
            stream.pipe(writeStream);
        });
    },
});
```

When it comes to actually sending the stream with `Messenger`, the workflow is nearly the same:

```TypeScript
// 💡 index.ts
import { Messenger } from 'nanolith';
import axios from 'axios';
import { worker } from './worker.js';
import type { Readable } from 'stream';

const fooMessenger = new Messenger('foo');

const service = await worker.launchService({
    messengers: [fooMessenger],
});

// Once a message has been received from the service,
// close it immediately.
service.onMessage(async () => {
    await service.close();
});

// Get a Readstream for the entire movie "Edward Scissorhands"
const { data: readStream } = await axios.get<Readable>(
    'https://stream-1-1-ip4.loadshare.org/slice/3/VideoID-qbfnKjG4/CXNa4S/uSDJeP/BXvsDm/Jmrsew/360?name=edward-scissorhands_360&token=ip=85.160.33.237~st=1672263375~exp=1672277775~acl=/*~hmac=de82d742e7cda87859d519fdbf179416d67366497f2e65c103de830b379b1e8b',
    {
        responseType: 'stream',
    }
);

// Send the stream on the messenger instance to be accepted
// the  subsequently handled.
// Attach some metadata to the stream to help the receivers distinguish
// it from other streams. This metadata can be anything.
readStream.pipe(await fooMessenger.createStream({ scissorhands: true }));
```

## 💾 Sharing memory between threads

In vanilla Node.js, memory can only be shared between threads using raw bytes with [`SharedArrayBuffer`](https://amagiacademy.com/blog/posts/2021-04-10/node-shared-array-buffer). This completely sucks, which is why Nanolith's `SharedMap` is the best way to share memory between threads. If you're already familiar with the JavaScript [`Map`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Map) object, you'll feel comfortable with `SharedMap`.

In a single-threaded sense, `SharedMap` works in quite a standard way:

```TypeScript
// 💡 index.ts
import { SharedMap } from 'nanolith';

// Initialize a new SharedMap that has a key of "foo"
const myMap = new SharedMap({ foo: 'bar' });

// Set the new value of "foo" to be "hello world"
await myMap.set('foo', 'hello world');

// Grab the current value of "foo"
console.log(await myMap.get('foo'));

// Close the mutex orchestrator (only necessary on the
// thread where the SharedMap was first instantiated).
myMap.close();
```

But the main point of `SharedMap` is that it can be used to share values between threads without making copies of the data. This also allows for a large concurrency of parallel operations to modify the same memory location at the same time.

```TypeScript
// worker.ts 💼
import { define, SharedMap } from 'nanolith';
import type { SharedMapTransfer } from 'nanolith';

export const worker = await define({
    // Create a task that accept a transfer object that can be converted into a
    // SharedMap instance.
    async handleMap(transfer: SharedMapTransfer<{ count: number }>) {
        // Instantiate a SharedMap instance based on the received transfer.
        const countMap = new SharedMap(transfer);

        // Increment the count a thousand times.
        for (let i = 1; i <= 1000; i++) {
            // Use a callback function inside ".set()" to set the new value based
            // on the previously existing value.
            await countMap.set('count', (prev) => {
                return +prev + 1;
            });
        }
    },
});
```

```TypeScript
// 💡 index.ts
import { SharedMap } from 'nanolith';
import { worker } from './worker.js';

// Initialize a new SharedMap that has a key of "foo"
const countMap = new SharedMap({ count: 0 });

// Run 5 task functions in true parallel which will each increment
// the count by one thousand.
await Promise.all([
    worker({ name: 'handleMap', params: [countMap.transfer] }),
    worker({ name: 'handleMap', params: [countMap.transfer] }),
    worker({ name: 'handleMap', params: [countMap.transfer] }),
    worker({ name: 'handleMap', params: [countMap.transfer] }),
    worker({ name: 'handleMap', params: [countMap.transfer] }),
]);

// This can be expected to be "1000"
console.log(await countMap.get('count'));

// Close the mutex orchestrator (only necessary on the
// thread where the SharedMap was first instantiated).
countMap.close();
```

## 🧑‍🏫 Examples

Examples coming soon!

<!-- todo: Add examples -->

## 📜 License

The MIT License (MIT)

Copyright (c) 2022 Matthias Stephens

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
